# /packages/intranet-ubl/tcl/intranet-ubl-procs.tcl
#
# Copyright (C) 2003-2007 ]project-open[
#
# All rights reserved. Please check
# http://www.project-open.com/license/ for details.

ad_library {
    Universal Business Language Export (no Import yet) Package
    @author frank.bergmann@project-open.com
    @creation-date 10 December 2007
    @cvs-id $Id$
}

namespace eval im_ubl {

    # ----------------------------------------------------------------------
    # Export all non-exported financial documents to a directory
    # ----------------------------------------------------------------------

    ad_proc -private export_all_invoices {} {
	Schedule procedure to export all new invoices to a directory.
	"New" is determined by the ubl_exported_p flag in the table im_costs.
    } {
        # SemP: Only allow one process to process...
        if {[nsv_incr im_ubl_export ubl_export_p] > 1} {
            nsv_incr im_ubl_export ubl_export_p -1
            return -2
        }
	
	set files_exported -1
        set error_p [catch {
            ns_log Notice "im_ubl::export_all_invoices: About to export invoices"
        } err_msg]

	    set files_exported [export_all_invoices2]
	
        # SemV: Release Semaphore
        nsv_incr im_ubl_export ubl_export_p -1
	
	if {$error_p} {
	    im_send_alert_to_system_owner "Error with im_ubl::export_all_invoices" $err_msg
	    ad_return_complaint 1 "Error in im_ubl::export_all_invoices:<br><pre>$err_msg</pre>"
	}
	
	return $files_exported
    }
    
    ad_proc -private export_all_invoices2 { } {
	Schedule procedure to export all new invoices to a directory.
        "New" is determined by the ubl_exported_p flag in the table im_costs.
    } {
	set export_path [parameter::get_from_package_key -package_key "intranet-ubl" -parameter "UblExportPathUnix" -default "/web/projop/filestorage/ubl"]
	set converter_proc [parameter::get_from_package_key -package_key "intranet-ubl" -parameter "UblExportConverterProc" -default "invoice2ubl"]
	set max_invoices [parameter::get_from_package_key -package_key "intranet-ubl" -parameter "UblExportMaxInvoices" -default 500]
    	
	set invoices_sql "
		select	i.invoice_id,
			c.cost_name
		from
			im_costs c,
			im_invoices i
		where
			c.cost_id = i.invoice_id
			and c.cost_type_id = [im_cost_type_invoice]
			and c.ubl_exported_p = 'f'
		order by invoice_id DESC
		LIMIT :max_invoices
        "
	set files_exported 0

	db_foreach export_invoices $invoices_sql {
	    set xml [eval $converter_proc -invoice_id $invoice_id]
	    set file_body "$cost_name.xml"
	    set file_name "$export_path/$file_body"
	    
	    if {[catch {
		set fd [open $file_name "w"]
		puts -nonewline $fd $xml
		close $fd
	    } err_msg]} {
		ad_return_complaint 1 "<b>Error creating file '$file_name'</b>:<br><pre>$err_msg</pre>"
		ad_script_abort
	    }
	    
	    if {"" != $xml} {
		db_dml invoice_written "
			update im_costs set 
				ubl_exported_p = 't',
				read_only_p = 't'
		 	where cost_id = :invoice_id
	        "
	    }
	    incr files_exported
	}

	return $files_exported
    }


    # ----------------------------------------------------------------------
    # Create an XML UBL structure for a single document
    # ----------------------------------------------------------------------
    
    ad_proc -private invoice2ubl_v10 { -invoice_id } {
	V1.0 version of the export. These "legacy" versions are designed for
	customers who don't want to change output.
    } {
	return [invoice2ubl -invoice_id $invoice_id]
    }
	
    ad_proc -private invoice2ubl {
	-invoice_id
    } {
	# rounding precision can be between 2 (USD,EUR, ...)
	# and -5 (Old Turkish Lira, ...).
	set rounding_precision 2
	set required_field "<font color=red size=+1><B>*</B></font>"
	set rounding_factor [expr exp(log(10) * $rounding_precision)]
	set rf $rounding_factor
	
	# US American number format
	set locale "en_US"
	
	set cur_format [im_l10n_sql_currency_format -locale $locale]
	set vat_format $cur_format
	set tax_format $cur_format
	
	# ---------------------------------------------------------------
	# Get information about the financial document ("invoice")
	# ---------------------------------------------------------------

	set buyer_id_sql [parameter::get_from_package_key -package_key "intranet-ubl" -parameter "BuyerIdSqlExpression" -default "cust.company_id"]
	set seller_id_sql [parameter::get_from_package_key -package_key "intranet-ubl" -parameter "SellerIdSqlExpression" -default "prov.company_id"]



	set customer_address_id_sql [parameter::get_from_package_key -package_key "intranet-ubl" -parameter "CustomerAddressIdSqlExpression" -default "ocust.office_id"]
	set provider_address_id_sql [parameter::get_from_package_key -package_key "intranet-ubl" -parameter "ProviderAddressIdSqlExpression" -default "oprov.office_id"]


	if {![db_0or1row project_info "
		select	c.*,
			i.*,
	                paymeth.category_description as payment_method_desc,
	
			$buyer_id_sql as buyer_id,
			$seller_id_sql as seller_id,

			cust.company_name as customer_name,
			cust.main_office_id as customer_main_office_id,
			i.invoice_office_id as invoice_office_id,
			coalesce(
				i.company_contact_id,
				cust.accounting_contact_id,
				cust.primary_contact_id
			) as cust_contact_id,
	
			prov.company_name as provider_name,
			prov.main_office_id as provider_main_office_id,
			coalesce(
				prov.accounting_contact_id,
				prov.primary_contact_id
			) as prov_contact_id,
	
			to_char(o.creation_date, 'YYYY-MM-DD') as issue_date_pretty
		from
			im_costs c,
			im_invoices i
	                LEFT OUTER JOIN im_categories paymeth ON (i.payment_method_id = paymeth.category_id),
			acs_objects o,
			im_companies cust,
			im_companies prov
		where
			i.invoice_id = :invoice_id
			and i.invoice_id = c.cost_id
			and i.invoice_id = o.object_id
			and c.customer_id = cust.company_id
			and c.provider_id = prov.company_id
        "]} {
	    ad_return_complaint 1 [lang::message::lookup "" intranet-ubl.Financial_Document_Not_Found "Didn't find financial document \#%invoice_id%"]
	    ad_script_abort
	}
	
	set cust_contact_sql "
		select
			im_name_from_user_id(uc.user_id) as customer_contact_name,
			uc.work_phone as customer_contact_work_phone
		from
			users_contact uc
		where
			uc.user_id = :cust_contact_id
        "
	set customer_contact_exists_p [db_0or1row cust_contact $cust_contact_sql]

	set prov_contact_sql "
		select
			im_name_from_user_id(uc.user_id) as provider_contact_name,
			uc.work_phone as provider_contact_work_phone
		from
			users_contact uc
		where
			uc.user_id = :prov_contact_id
        "
	set provider_contact_exists_p [db_0or1row prov_contact $prov_contact_sql]


	# Office Logic: Take the customers and provider's main office,
	# unless there has been a specific delivery office specified for the customer.
	set cust_office_id $customer_main_office_id
	set prov_office_id $provider_main_office_id
	if {"" != $invoice_office_id} { set cust_office_id $invoice_office_id}
	
	set prov_office_sql "
		select
			$provider_address_id_sql as provider_office_id,
			oprov.address_line1 as provider_address_line1,
			oprov.address_line2 as provider_address_line2,
			oprov.address_city as provider_address_city,
			oprov.address_state as provider_address_state,
			oprov.address_postal_code as provider_address_postal_code,
			upper(oprov.address_country_code) as provider_address_country_code,
			prov_countries.country_name as provider_address_country_name
		from
			im_offices oprov
			LEFT OUTER JOIN country_codes prov_countries 
				ON (oprov.address_country_code = prov_countries.iso)
		where
			oprov.office_id = :prov_office_id
        "
	if {![db_0or1row prov_office $prov_office_sql]} {
	    ad_return_complaint 1 [lang::message::lookup "" intranet-ubl.Provider_Office_for_Financial_Document_Not_Found "Didn't find Provider's office for financial document \#%invoice_id%"]
	    ad_script_abort
	}
	
	set cust_office_sql "
		select
			$customer_address_id_sql as customer_office_id,
			ocust.address_line1 as customer_address_line1,
			ocust.address_line2 as customer_address_line2,
			ocust.address_city as customer_address_city,
			ocust.address_state as customer_address_state,
			ocust.address_postal_code as customer_address_postal_code,
			upper(ocust.address_country_code) as customer_address_country_code,
			cust_countries.country_name as customer_address_country_name
		from
			im_offices ocust
			LEFT OUTER JOIN country_codes cust_countries 
				ON (ocust.address_country_code = cust_countries.iso)
		where
			ocust.office_id = :cust_office_id
        "
	if {![db_0or1row cust_office $cust_office_sql]} {
	    ad_return_complaint 1 [lang::message::lookup "" intranet-ubl.Customer_Office_for_Financial_Document_Not_Found "Didn't find customer office for financial document \#%invoice_id%"]
	    ad_script_abort
	}
	

	# Calculate grand total based on the same inner SQL
	db_1row calc_grand_total "
		select	i.*,
			round(i.grand_total * :vat / 100 * :rf) / :rf as vat_amount,
			round(i.grand_total * :tax / 100 * :rf) / :rf as tax_amount,
			i.grand_total
				+ round(i.grand_total * :vat / 100 * :rf) / :rf
				+ round(i.grand_total * :tax / 100 * :rf) / :rf
			as total_due
		from
			(select
				max(i.currency) as currency,
				sum(i.amount) as subtotal,
				round(sum(i.amount) * :surcharge_perc::numeric) / 100.0 as surcharge_amount,
				round(sum(i.amount) * :discount_perc::numeric) / 100.0 as discount_amount,
				sum(i.amount)
					+ round(sum(i.amount) * :surcharge_perc::numeric) / 100.0
					+ round(sum(i.amount) * :discount_perc::numeric) / 100.0
				as grand_total
			from 
				(select	ii.*,
					round(ii.price_per_unit * ii.item_units * :rf) / :rf as amount
				from	im_invoice_items ii,
					im_invoices i
				where	i.invoice_id = ii.invoice_id
					and i.invoice_id = :invoice_id
				) i
			) i
        "
	
	set subtotal_pretty [lc_numeric [im_numeric_add_trailing_zeros [expr $subtotal+0] $rounding_precision] "" $locale]
	set vat_amount_pretty [lc_numeric [im_numeric_add_trailing_zeros [expr $vat_amount+0] $rounding_precision] "" $locale]
	set tax_amount_pretty [lc_numeric [im_numeric_add_trailing_zeros [expr $tax_amount+0] $rounding_precision] "" $locale]
	set grand_total_pretty [lc_numeric [im_numeric_add_trailing_zeros [expr $grand_total+0] $rounding_precision] "" $locale]
	set total_due_pretty [lc_numeric [im_numeric_add_trailing_zeros [expr $total_due+0] $rounding_precision] "" $locale]
	set discount_perc_pretty $discount_perc
	set surcharge_perc_pretty $surcharge_perc
	
	
	# Get the right outer tag name, depending on the document type
	set document_type "error"
	switch $cost_type_id {
	    3700 { set document_type "Invoice" }
	    3702 { set document_type "Quote" }
	    3704 { set document_type "Invoice" }
	    3706 { set document_type "Order" }
	    3724 { set document_type "DelNote" }
	}

	# ---------------------------------------------------------------
	# Create the XML
	# ---------------------------------------------------------------

	set version "1.12"
	set view_index 0
	
	# ---------------------------------------------------------------
	# Main Document Node
	
	set doc [dom createDocument $document_type]
	set main_node [$doc documentElement]
	
	$main_node setAttribute xmlns:cbc "urn:oasis:names:specification:ubl:schema:xsd:CommonBasicComponents-1.0" 
	$main_node setAttribute	xmlns:cac "urn:oasis:names:specification:ubl:schema:xsd:CommonAggregateComponents-1.0" 
	$main_node setAttribute	xmlns:cur "urn:oasis:names:specification:ubl:schema:xsd:CurrencyCode-1.0" 
	$main_node setAttribute	xmlns:xsi "http://www.w3.org/2001/XMLSchema-instance" 
	$main_node setAttribute	xmlns "urn:oasis:names:specification:ubl:schema:xsd:Invoice-1.0" 
	$main_node setAttribute	xsi:schemaLocation "urn:oasis:names:specification:ubl:schema:xsd:Invoice-1.0 ../../xsd/maindoc/UBL-Invoice-1.0.xsd"
	
	$main_node appendXML "<ID>$cost_name</ID>"
	$main_node appendXML "<cbc:IssueDate>$issue_date_pretty</cbc:IssueDate>"
	$main_node appendXML "<TaxPointDate>$issue_date_pretty</TaxPointDate>"

	$main_node appendXML "	
	<cac:OrderReference>
		<cac:BuyersID>$buyer_id</cac:BuyersID>
		<cac:SellersID>$seller_id</cac:SellersID>
		<cbc:IssueDate>$issue_date_pretty</cbc:IssueDate>
	</cac:OrderReference>"

	set customer_contact_xml ""
	if {$customer_contact_exists_p} {
	    set customer_contact_xml "
		<cac:AccountsContact>
			<cbc:Name>$customer_contact_name</cbc:Name>
			<cbc:Telephone>$customer_contact_work_phone</cbc:Telephone>
		</cac:AccountsContact>
            "
	}

	$main_node appendXML "
	<cac:BuyerParty>
		<cac:Party>
			<cac:PartyName>
				<cbc:Name>$customer_name</cbc:Name>
			</cac:PartyName>
			<cac:Address>
				<ID>$customer_office_id</ID>
				<cbc:StreetName>$customer_address_line1 $customer_address_line2</cbc:StreetName>
				<cbc:CityName>$customer_address_city</cbc:CityName>
				<cbc:PostalZone>$customer_address_postal_code</cbc:PostalZone>
				<cac:Country>
					<cac:IdentificationCode>$customer_address_country_code</cac:IdentificationCode>
					<cac:Name>$customer_address_country_name</cac:Name>
				</cac:Country>
				<cbc:CountrySubentityCode>$customer_address_state</cbc:CountrySubentityCode>
			</cac:Address>
		</cac:Party>
		$customer_contact_xml
	</cac:BuyerParty>"


        set provider_contact_xml ""
        if {$provider_contact_exists_p} {
            set provider_contact_xml "
		<cac:AccountsContact>
			<cbc:Name>$provider_contact_name</cbc:Name>
			<cbc:Telephone>$provider_contact_work_phone</cbc:Telephone>
		</cac:AccountsContact>
            "
        }


	$main_node appendXML "
	<cac:SellerParty>
		<cac:Party>
			<cac:PartyName>
				<cbc:Name>$provider_name</cbc:Name>
			</cac:PartyName>
			<cac:Address>
				<ID>$provider_office_id</ID>
				<cbc:StreetName>$provider_address_line1 $provider_address_line2</cbc:StreetName>
				<cbc:CityName>$provider_address_city</cbc:CityName>
				<cbc:PostalZone>$provider_address_postal_code</cbc:PostalZone>
				<cac:Country>
					<cac:IdentificationCode>$provider_address_country_code</cac:IdentificationCode>
					<cac:Name>$provider_address_country_name</cac:Name>
				</cac:Country>
				<cbc:CountrySubentityCode>$provider_address_state</cbc:CountrySubentityCode>
			</cac:Address>
		</cac:Party>
		$provider_contact_xml
	</cac:SellerParty>"


	# Filter out any HTML tags from payment method and replace with a space
	regsub -all {\<[^\>]*\>} $payment_method_desc " " payment_method_desc
	
	$main_node appendXML "
	<cac:PaymentTerms>
		<cbc:Note>$payment_method_desc</cbc:Note>
		<cac:ReferenceEventCode>!</cac:ReferenceEventCode>
	</cac:PaymentTerms>"

	$main_node appendXML "
	<cac:TaxTotal>
		<cbc:TotalTaxAmount amountCurrencyCodeListVersionID=\"0.3\" amountCurrencyID=\"$currency\">$vat_amount_pretty</cbc:TotalTaxAmount>
		<cac:TaxSubTotal>
			<cbc:TaxableAmount amountCurrencyCodeListVersionID=\"0.3\" amountCurrencyID=\"$currency\">$amount</cbc:TaxableAmount>
			<cbc:TaxAmount amountCurrencyCodeListVersionID=\"0.3\" amountCurrencyID=\"$currency\">$vat_amount_pretty</cbc:TaxAmount>
			<cac:TaxCategory>
				<cac:ID>STD</cac:ID>
				<cbc:Percent>$vat</cbc:Percent>
				<cac:TaxScheme>
					<cac:TaxTypeCode>VAT</cac:TaxTypeCode>
				</cac:TaxScheme>
			</cac:TaxCategory>
		</cac:TaxSubTotal>
	</cac:TaxTotal>"

	$main_node appendXML "
	<cac:LegalTotal>
		<cbc:LineExtensionTotalAmount amountCurrencyCodeListVersionID=\"0.3\" 
			amountCurrencyID=\"$currency\">$subtotal_pretty</cbc:LineExtensionTotalAmount>
		<cbc:TaxExclusiveTotalAmount amountCurrencyCodeListVersionID=\"0.3\" 
			amountCurrencyID=\"$currency\">$subtotal_pretty</cbc:TaxExclusiveTotalAmount>
		<cbc:TaxInclusiveTotalAmount amountCurrencyCodeListVersionID=\"0.3\" 
			amountCurrencyID=\"$currency\">$grand_total_pretty</cbc:TaxInclusiveTotalAmount>
	</cac:LegalTotal>"


	set invoice_items_sql "
		select
			i.*,
			i.item_id as invoice_item_id,
			p.*,
			im_category_from_id(i.item_type_id) as item_type,
			im_category_from_id(i.item_uom_id) as item_uom,
			p.project_nr as project_short_name,
			round(i.price_per_unit * i.item_units * :rf) / :rf as amount,
			to_char(round(i.price_per_unit * i.item_units * :rf) / :rf, :cur_format) as amount_formatted
		from
			im_invoice_items i
			LEFT JOIN im_projects p on i.project_id=p.project_id
		where
			i.invoice_id=:invoice_id
		order by
			i.sort_order,
			i.item_type_id;
        "

	set ctr 1
	db_foreach invoice_items $invoice_items_sql {
	    
	    set amount_pretty [lc_numeric [im_numeric_add_trailing_zeros [expr $amount+0] $rounding_precision] "" $locale]
	    set item_units_pretty [lc_numeric [expr $item_units+0] "" $locale]
	    set price_per_unit_pretty [lc_numeric [im_numeric_add_trailing_zeros [expr $price_per_unit+0] $rounding_precision] "" $locale]
	    
	    $main_node appendXML "
	<cac:${document_type}Line>
		<cac:ID>$ctr</cac:ID>
		<cbc:InvoicedQuantity quantityUnitCode=\"$item_uom\">$item_units_pretty</cbc:InvoicedQuantity>
		<cbc:LineExtensionAmount amountCurrencyCodeListVersionID=\"0.3\" 
			amountCurrencyID=\"$currency\">[string trim $amount_pretty]</cbc:LineExtensionAmount>
		<cac:OrderLineReference>
			<cac:BuyersLineID>$ctr</cac:BuyersLineID>
		</cac:OrderLineReference>
		<cac:Item>
			<cbc:Description>[ns_quotehtml $item_name]</cbc:Description>
			<cac:BuyersItemIdentification>
				<cac:ID>$invoice_item_id</cac:ID>
			</cac:BuyersItemIdentification>
			<cac:SellersItemIdentification>
				<cac:ID>$invoice_item_id</cac:ID>
			</cac:SellersItemIdentification>
			<cac:BasePrice>
				<cbc:PriceAmount amountCurrencyCodeListVersionID=\"0.3\" 
					amountCurrencyID=\"$currency\">$price_per_unit_pretty
				</cbc:PriceAmount>
			</cac:BasePrice>
		</cac:Item>
	</cac:${document_type}Line>"

	    incr ctr
	}
	return [$doc asXML -indent 2]
    }
    
    # End of library
}


